package tbox;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;

import org.w3c.dom.*;
import org.xml.sax.*;

import domain.*;
import assertions.*;

public class TBoxParser {
	private TBox tbox;
	private AtomicConcept topC;
	private AtomicValueDomain topD;
	private NodeList rootChildren = null;
	
	public TBoxParser(String xmlPath) throws ParserConfigurationException, SAXException, IOException {
	
		Document xmltbox = parseXML(xmlPath);
		
		topC = new AtomicConcept("TopC");
		topD = new AtomicValueDomain("TopD");
		
		tbox = new TBox();
		
		boolean alphabetFound = false;
		boolean tboxFound = false;
		rootChildren = xmltbox.getChildNodes();
		for (int i = 0; i < rootChildren.getLength(); i++) {
			Node act = rootChildren.item(i);
			if (act.getNodeName().equalsIgnoreCase("ontology")) {
				NodeList under = act.getChildNodes();
				
				for (int j=0; j < under.getLength(); j++) {
					if (under.item(j).getNodeName().equalsIgnoreCase("alphabet")) {
						System.out.print("Parsing TBox alphabet...");
						alphabetFound = true;
						loadAlphabet(under.item(j));
						System.out.println(" DONE!");
					}
				}
				
				for (int j=0; j < under.getLength(); j++) {
					if (under.item(j).getNodeName().equalsIgnoreCase("tbox")) {
						System.out.print("Parsing TBox assertions...");
						tboxFound = true;
						loadTBox(under.item(j));
						System.out.println(" DONE!");
					}
				}
			}
		}
		if (!alphabetFound)
			throw new RuntimeException("Can't load TBox alphabet!");
		if (!tboxFound)
			throw new RuntimeException("Can't load TBox assertions!");
	}
	
	
	private Document parseXML(String xmlFileName) throws ParserConfigurationException, SAXException, IOException {

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setValidating(true);
		factory.setIgnoringElementContentWhitespace(true);
		factory.setIgnoringComments(true);
		DocumentBuilder builder = null;
		Document document = null;

		builder = factory.newDocumentBuilder();
		document = builder.parse(xmlFileName);

		return document;
	}
	
	private void loadAlphabet(Node node) {
		
		NodeList list = node.getChildNodes();
		for (int i = 0; i < list.getLength(); ++i) {
			Node nameNode = list.item(i);
			String nameType = nameNode.getNodeName();
			String name = nameNode.getFirstChild().getNodeValue();
			if(nameType.equals("atomicC")) {
				AtomicConcept concept = new AtomicConcept(name);
				tbox.getAlfabeth().add(concept);
			}

			else if(nameType.equals("atomicR")) {
				AtomicRole role = new AtomicRole(name);
				tbox.getAlfabeth().add(role);
			}

			else if(nameType.equals("atomicV")) {
				AtomicValueDomain valueSet = new AtomicValueDomain(name);
				tbox.getAlfabeth().add(valueSet);
			}

			else if(nameType.equals("atomicCA")) {
				AtomicConceptAttribute conceptAttr = new AtomicConceptAttribute(name);
				tbox.getAlfabeth().add(conceptAttr);
			}

			else if(nameType.equals("atomicRA")) {
				AtomicRoleAttribute roleAttr = new AtomicRoleAttribute(name);
				tbox.getAlfabeth().add(roleAttr);
			}
		}
	}

	private void loadTBox(Node node) {
		
		NodeList list = node.getChildNodes();
		for (int i = 0; i < list.getLength(); ++i) {
			String assertion = list.item(i).getNodeName();
			if (assertion.equals("inclusionAssertion")) {
				NodeList asse = list.item(i).getChildNodes();
				Node subsumee = asse.item(0);
				if(subsumee.getNodeName().equals("basicC"))
					loadConceptInclusion(list.item(i));
				else if(subsumee.getNodeName().equals("basicR"))
					loadRoleInclusion(list.item(i));
				else if(subsumee.getNodeName().equals("basicV"))
					loadValueSetInclusion(list.item(i));
				else if(subsumee.getNodeName().equals("atomicCA"))
					loadConceptAttrInclusion(list.item(i));
				else //atomicRA
					loadRoleAttrInclusion(list.item(i));
			}
			else if(assertion.equals("funct")) {
				Node functObj = list.item(i).getFirstChild();
				if(functObj.getNodeName().equals("basicR")) {
					BasicRole role = createBasicRole(functObj);
					tbox.addAssertion(new FunctionalityAssertion(role));
				}
				else if(functObj.getNodeName().equals("atomicCA")) {
					String conceptAttrName = functObj.getFirstChild().getNodeValue();
					AtomicConceptAttribute atomicCA = new AtomicConceptAttribute(conceptAttrName);
					tbox.addAssertion(new FunctionalityAssertion(atomicCA));
				}
				else {
					String roleAttrName = functObj.getFirstChild().getNodeValue();
					AtomicRoleAttribute atomicRA = new AtomicRoleAttribute(roleAttrName);
					tbox.addAssertion(new FunctionalityAssertion(atomicRA));
				}
			}
			else if(assertion.equals("IdC")) {
				/*non vengono trattate per il momento le identification assertion
				 * in quanto non presenti nella DTD di quonto
				 */
				//loadIdentification(list.item(i));
			}
		}
	}
	
	private void loadConceptInclusion(Node node) {

		NodeList list = node.getChildNodes();
		Node basicC = list.item(0);
		Included includedC = createBasicConcept(basicC);

		for(int i = 1; i < list.getLength(); ++i) {
			Including includingC = createGeneralConcept(list.item(i));
			tbox.addAssertion(new InclusionAssertion(includedC, includingC));
		}
	}
	
	private void loadRoleInclusion(Node node)  {

		NodeList list = node.getChildNodes();
		Node basicR = list.item(0);
		Included includedR = createBasicRole(basicR);

		for(int i = 1; i < list.getLength(); ++i) {
			Including includingR = createGeneralRole(list.item(i));
			//if(isAtomicRole(list.item(i)))	checkRoleFunct((AtomicRole)includingR); lancia eccezzione
			tbox.addAssertion(new InclusionAssertion(includedR, includingR));
		}
	}
	
	private void loadValueSetInclusion(Node node) {

		NodeList list = node.getChildNodes();
		Node basicV = list.item(0);
		Included includedV = createBasicValueDomain(basicV);

		for(int i = 1; i < list.getLength(); ++i) {
			GeneralValueDomain includingV = createGeneralValueDomain(list.item(i));
			tbox.addAssertion(new InclusionAssertion(includedV, includingV));
			if (includingV instanceof AtomicValueDomain) {
				AtomicValueDomain t = (AtomicValueDomain)includingV;
				tbox.getAlfabeth().add(t);
			}
		}
	}
	
	private void loadConceptAttrInclusion(Node node) {

		NodeList list = node.getChildNodes();
		Node atomicCA = list.item(0);
		String attrName = atomicCA.getFirstChild().getNodeValue();
		Included includedCA = new AtomicConceptAttribute(attrName);

		for(int i = 1; i < list.getLength(); ++i)	{
			Including includingCA = createGeneralConceptAttribute(list.item(i));
			//if(isAtomicAttribute(list.item(i)))	checkCAttrFunct((ConceptAttribute)includingCA);
			tbox.addAssertion(new InclusionAssertion(includedCA, includingCA));
		}
	}
	
	private void loadRoleAttrInclusion(Node node) {

		NodeList list = node.getChildNodes();
		Node atomicRA = list.item(0);
		String attrName = atomicRA.getFirstChild().getNodeValue();
		Included includedRA = new AtomicRoleAttribute(attrName);

		for(int i = 1; i < list.getLength(); ++i) {
			Including includingRA = createGeneralRoleAttribute(list.item(i));
			//if(isAtomicAttribute(list.item(i)))	checkRAttrFunct((RoleAttribute)includingRA);
			tbox.addAssertion(new InclusionAssertion(includedRA, includingRA));
		}
	}
	
	private BasicConcept createBasicConcept(Node basicC) {

		BasicConcept basicConcept = null;
		Node basicCType = basicC.getFirstChild();

		if(basicCType.getNodeName().equals("atomicC")) {
			String atomicCName = basicCType.getFirstChild().getNodeValue();
			basicConcept = new AtomicConcept(atomicCName);
		}
		else if(basicCType.getNodeName().equals("exists")) {
			Node basicR = basicCType.getFirstChild();
			BasicRole basicRole = createBasicRole(basicR);
			basicConcept = new ExistentialRole(basicRole);
		}
		else if(basicCType.getNodeName().equals("CADomain")) {
			String conceptAttrName = basicCType.getFirstChild().getFirstChild().getNodeValue();
			AtomicConceptAttribute atomicCA = new AtomicConceptAttribute(conceptAttrName);
			basicConcept = new ConceptAttributeDomain(atomicCA);
		}
		return basicConcept;
	}
	
	private BasicRole createBasicRole(Node basicR) {

		BasicRole basicRole = null;
		Node basicRType = basicR.getFirstChild();

		if(basicRType.getNodeName().equals("atomicR")) {

			String atomicRName = basicRType.getFirstChild().getNodeValue();
			AtomicRole atomicRole = new AtomicRole(atomicRName);
			if (!isInverse(basicR)) basicRole = new DirectRole(atomicRole);
			else basicRole = new InverseRole(atomicRole);
		}
		else {
			String roleAttrName = basicRType.getFirstChild().getFirstChild().getNodeValue();
			AtomicRoleAttribute atomicRA = new AtomicRoleAttribute(roleAttrName);
			if (!isInverse(basicR)) basicRole = new DirectRoleAttributeDomain(atomicRA);
			else basicRole =  new InverseRoleAttributeDomain(atomicRA);
		}
		return basicRole;
	}
	
	private BasicValueDomain createBasicValueDomain(Node basicV) {
		/* NOTA_ modifica della DTD per seguire la nuova grammatica
		 * BasicValueDomain pu essere solo un range, non pu essere un AtomicValueDomain
		 */
		BasicValueDomain basicValueDomain = null;
		Node basicVType = basicV.getFirstChild();

		if(basicVType.getNodeName().equals("ARange")) {
			Node atomicAttr = basicVType.getFirstChild();
			String atomicAttrName = atomicAttr.getFirstChild().getNodeValue();
			if(atomicAttr.getNodeName().equals("atomicCA"))
				basicValueDomain = new ConceptAttributeRange(new AtomicConceptAttribute(atomicAttrName));
			else
				basicValueDomain = new RoleAttributeRange(new AtomicRoleAttribute(atomicAttrName));
		}
		return basicValueDomain;
	}
	
	private GeneralConceptAttribute createGeneralConceptAttribute(Node generalCA) {

		GeneralConceptAttribute generalConceptAttr = null;
		String conceptAttrName = generalCA.getFirstChild().getFirstChild().getNodeValue();
		AtomicConceptAttribute atomicCA = new AtomicConceptAttribute(conceptAttrName);
		
		if(((Element)generalCA).getAttribute("sign").equals("positive"))
			generalConceptAttr = atomicCA;
		else
			generalConceptAttr = new NegatedConceptAttribute(atomicCA);

		return generalConceptAttr;
	}

	private GeneralRoleAttribute createGeneralRoleAttribute(Node generalRA) {

		GeneralRoleAttribute generalRoleAttr = null;
		String roleAttrName = generalRA.getFirstChild().getFirstChild().getNodeValue();
		AtomicRoleAttribute atomicRA = new AtomicRoleAttribute(roleAttrName);
		
		if(((Element)generalRA).getAttribute("sign").equals("positive"))
			generalRoleAttr = atomicRA;
		else
			generalRoleAttr = new NegatedRoleAttribute(atomicRA);

		return generalRoleAttr;
	}
	
	private GeneralConcept createGeneralConcept(Node generalC) {

		GeneralConcept generalConcept = null;
		Node generalCType = generalC.getFirstChild();
		
		if(generalCType.getNodeName().equals("topC"))
			generalConcept = topC;
		
		else if(generalCType.getNodeName().equals("signedC")) {
			if(((Element)generalCType).getAttribute("sign").equals("positive"))
				generalConcept = createBasicConcept(generalCType.getFirstChild());
			else
				generalConcept = new NegatedBasicConcept(createBasicConcept(generalCType.getFirstChild()));
		}
		return generalConcept;
	}
	
	private GeneralRole createGeneralRole(Node generalR) {

		GeneralRole generalRole = null;
		Node generalRType = generalR.getFirstChild();

		if(generalRType.getNodeName().equals("signedR")) {
			if(((Element)generalRType).getAttribute("sign").equals("positive"))
				generalRole = createBasicRole(generalRType.getFirstChild());
			else
				generalRole = new NegatedBasicRole(createBasicRole(generalRType.getFirstChild()));
		}
		return generalRole;
	}
	
	private GeneralValueDomain createGeneralValueDomain(Node generalV) {

		GeneralValueDomain generalValueDomain = null;
		Node valueSetType = generalV.getFirstChild();
		String name = valueSetType.getFirstChild().getNodeValue();
		String datatype = name.substring(3).toUpperCase();

		if(valueSetType.getNodeName().equals("topD"))
			generalValueDomain = topD;
		
		else if(valueSetType.getNodeName().equals("predefinedV"))
			generalValueDomain = new AtomicValueDomain(datatype);
		
		return generalValueDomain;
	}
	
	
	private boolean isInverse(Node node) {

		boolean inverse;
		if(((Element)node).getAttribute("dir").equals("direct"))
			inverse = false;
		else
			inverse = true;

		return inverse;
	}
	
	public TBox getTbox() {
		return tbox;
	}
	
	
}
